module ScannerConnection
  class ScannerConnectionWrapper
    class InvalidLogLevel < StandardError; end

    def initialize(ip, port)
      @client = nil
      @ip = ip
      @port = port

      Kernel.at_exit do
        @client.close unless @client.nil?
      end
    end

    def ping
      ensure_connection
      close_client
      {success: true}
    rescue => err
      close_client
      log("failed to ping: #{err}", :warn)
      {success: false, error: err}
    end

    def send_message(message)
      ensure_connection
      @client.write(message)
    end

    def read_response(number_of_bytes)
      ensure_connection
      @client.recvfrom(number_of_bytes)
    end

    def ensure_connection
      require "socket"
      if @client.nil?
        log("starting connection.")
        open_client
      elsif @client.remote_address.ip_address != @ip
        log("remote ip invalid, reconnecting.")
        restart_client
      end
    rescue IOError => err
      log("IO error: #{err}.", :warn)
      restart_client
    end

    def close
      close_client
    end

    private

    def open_client
      @client = Socket.tcp(@ip, @port, connect_timeout: 2)
      log "connection opened. remote address: #{@client.remote_address.ip_address}:#{@client.remote_address.ip_port}."
    end

    def close_client
      log("closing client.")
      begin
        @client.close unless @client.nil?
        @client = nil
      rescue
        # close can throw for a number of reason, let it be since the connection will be dead regardless.
        # forcing the client to close will also consequently error out any active call that could be holding up the mutex
      end
    end

    def restart_client
      close_client
      open_client
    end

    def log(message, level=:info)
      raise InvalidLogLevel unless [:info, :warn, :error].include?(level)
      Rails.logger.send(level, "#{@ip}:#{@port} #{message}")
    end
  end
end